#!/usr/bin/python

from __future__ import division

import sys
import os
import lzo
import time
import zlib
import dioutil


BLOCK_SIZE = 4 * 1024 * 1024

class Restore_Engine(object):
	def __init__(self, client):
		self._status_file = None
		
		self._index_new = None
		self._index_old = None
		
		self._client = client
		
		self._dst = None
		self._dst_read = None
		self._dst_lv = None
		
		self._backup_identifier = None
		
		self._last_block = None
		self._block_range = []
		
		self._run = False
		self._last_activity = 0
		
		self._job_id = None
		self._result = [0, 'init']
		# result - [error_code, text_status]
		# 0, status - OK/Cancel
		# 1, breaking - backup break correctly
		# 2, read_end - connection lost
		# 3, io_error
		# 4, invalid_index - invalid backup data
		
	def _write_status(self, progress, block_number):
		fh = open(self._status_file, 'w')
		fh.write("%d\n" % int(block_number * BLOCK_SIZE))
		fh.close()
		self._client.update_restore_progress(self._backup_identifier, self._job_id, [block_number, self._last_block])
		self._last_activity = time.time()
		if progress != None:
			progress.set(float(block_number) / float(self._last_block))
			
	def start_restore(self, name, dst_lv, backup_identifier, job_id = None, disallow_discard = False):
		if(self._run):
			raise Exception('Restore is running')
		
		if(job_id == None):
			self._job_id = os.urandom(8).encode('hex')
			
		else:
			self._job_id = job_id
			
		self._status_file = '/dev/shm/vps-backup-process-status-%s' % (self._job_id,)
		
		self._name = name
		self._dst_lv = dst_lv
		self._backup_identifier = backup_identifier
		self._dst = dioutil.DioFile(disallow_discard = disallow_discard)
		self._dst.discard_zeroes = True
		self._dst.open(self._dst_lv, os.O_WRONLY)
		self._dst.seek(0)
		
		self._dst_read = dioutil.DioFile()
		self._dst_read.open(self._dst_lv)
		self._dst_read.seek(0)
				
		self._index_old = self._client.startRestore(self._name, self._backup_identifier, self._job_id)
		self._last_activity = time.time()
		self._index_new = {}
		
		if(len(self._index_old) == 0):
			self._result = [4, 'invalid_index_data']
			raise ValueError('Invalid index data')
		
		self._last_block = max(self._index_old.keys())
		self._block_range = range(self._last_block + 1)
		
		self._result = [0, 'restore_prepared']
		
		self._run = False
		
		return True
	
	def resume_restore(self, name = None, dst_lv = None, backup_identifier = None, index = None, client = None, job_id = None, disallow_discard = False):
		if(self._run):
			raise Exception('Restore is running')
			
		if(name != None):
			self._name = name
		
		if(dst_lv != None):
			self._dst_lv = dst_lv
			
		if(backup_identifier != None):
			self._backup_identifier = backup_identifier
			
		if(index != None):
			self._index_new = index
			
		if(client != None):
			self._client = client
			
		if(job_id != None):
			self._job_id = job_id
			
		if(self._name == None):
			raise Exception('Unknown backup source name')
			
		if(self._dst_lv == None):
			raise Exception('Unknown backup destination')
			
		if(self._backup_identifier == None):
			raise Exception('Unknown backup for continue')
			
		if(self._index_new == None):
			raise Exception('Unknown previous backup status')
			
		if(self._client == None):
			raise Exception('Backup client is not set')
		
		if(self._job_id == None):
			raise Exception('Job ID is not set')
			
		
		self._result = [0, 'restore_resume']	
		self._status_file = '/dev/shm/vps-backup-process-status-%s' % (self._job_id,)
		self._dst = dioutil.DioFile(disallow_discard = disallow_discard)
		self._dst.discard_zeroes = True
		self._dst.open(self._dst_lv, os.O_WRONLY)
		self._dst.seek((len(self._index_new) - 1) * BLOCK_SIZE)

		self._dst_read = dioutil.DioFile()
		self._dst_read.open(self._dst_lv)
		self._dst_read.seek((len(self._index_new) - 1) * BLOCK_SIZE)
		
		self._index_old = self._client.startRestore(self._name, self._backup_identifier, self._job_id)
		self._last_activity = time.time()
		
		self._last_block = max(self._index_old.keys())
		self._block_range = range(len(self._index_new) - 1, self._last_block + 1)
	
		if(len(self._index_old) == 0):
			self._result = [4, 'invalid_index_data']
			raise ValueError('Invalid index data')
		
		self._result = [0, 'restore_prepared']
		
		self._run = False
		
		return True
		
	def stop(self):
		self._run = False
		return True
		
	def run(self, compare = True, progress = None):
		if(self._run):
			raise Exception('Couldn\'t run restore twice')
		
		zero_block = chr(0) * BLOCK_SIZE
		change_count = 0
		
		self._run = True
		
		self._result = [0, 'restore_start']
		for block_number in self._block_range:
			if(not self._run):
				self._result = [1, 'breaking']
				return

			if(block_number % 10 == 0):
				try:
					self._write_status(progress, block_number)
					
				except:
					self._result = [2, 'connection_lost']
					self._run = False
					return
				
			if(compare):
				dst_block = self._dst_read.read(BLOCK_SIZE)
				if(not dst_block):
					self._dst.close()
					self._dst_read.close()
					self._result = [3, 'read_error']
					self._run = False
					return
				
				self._index_new[block_number] = [0, 0.0, '']
				if(block_number not in self._index_old):
					self._dst.close()
					self._dst_read.close()
					self._result = [4, 'invalid_index_data']
					self._run = False
					return
					
				if(dst_block == zero_block):
					self._index_new[block_number][0] = -1
					
				else:
					self._index_new[block_number][0] = lzo.adler32(dst_block)
					
				if(self._index_new[block_number][0] == self._index_old[block_number][0]):
					if(self._last_activity + 5 < time.time()):
						self._client.ping()
						self._last_activity = time.time()
						
					continue
					
				self._dst.seek(block_number * BLOCK_SIZE, 0)
				change_count += 1
				if(block_number > 300 and ((change_count * 100) / block_number) > 300):
					self._dst_read.close()
					compare = False
			
			if(self._index_old[block_number][0] == -1):
				self._dst.zero(BLOCK_SIZE)

			else:
				try:
					src_block = self._client.readBlock(self._name, self._backup_identifier, block_number)
					self._last_activity = time.time()
					if(self._index_old[block_number][2] == 'zlib'):
						data = zlib.decompress(src_block)
						self._dst.write(data)
					
					else:
						data = lzo.decompress(src_block)
						self._dst.write(data)
					
				except:
					self._dst.close()
					if(compare):
						self._dst_read.close()
						
					self._result = [3, 'write_error']
					self._run = False
					return
					
			if(self._last_activity + 5 < time.time()):
				self._client.ping()
				self._last_activity = time.time()
				
		self._dst.close()
		if(compare):
			self._dst_read.close()
		
		self._client.finishRestore(self._name, self._backup_identifier, self._job_id)
		
		try:
			os.unlink(self._status_file)
			
		except:
			pass
			
		return True
		
	def get_index_new(self):
		return self._index_new
		
	def get_result(self):
		return self._result
		
	def get_restore_status(self):
		return [len(self._index_new) if(self._index_new != None) else None, (self._last_block + 1) if(self._last_block != None) else None]

