import sys
import sqlite3, os, requests, pickle, subprocess, re, threading, time

from c2padmin import logger as logmodule

STRUCTURE_VERSION = 1
DB_STRUCTURE = ['''
	CREATE TABLE IF NOT EXISTS version (
		version INT
	)''',
	'''
	CREATE TABLE IF NOT EXISTS isos (
		url VARCHAR(255),
		size INT,
		last_modified VARCHAR(255),
		status VARCHAR(15),
		path VARCHAR(255),
		last_used TIMESTAMP DEFAULT CURRENT_TIMESTAMP
	)
	''',
	'''
	CREATE TABLE IF NOT EXISTS config (
		name VARCHAR(255) PRIMARY KEY,
		last_modified TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
		dirty INT DEFAULT 0
	)
	''']

logger = logmodule.getLogger(__name__)
import c2padmin

class LocalCache(object):
	"""docstring for LocalCache"""
	def __init__(self,cache_dir='/var/cache/c2padmin/'):
		self._cache_dir = cache_dir
		self._db_path = os.path.join(cache_dir,'db.sqlite')
		self._isos_dir = os.path.join(cache_dir,'isos')
		self._conf_dir = os.path.join(cache_dir,'conf')
		if not os.path.exists(self._cache_dir):
			os.mkdir(self._cache_dir)
			
		if not os.path.exists(self._isos_dir):
			os.mkdir(self._isos_dir)
			
		if not os.path.exists(self._conf_dir):
			os.mkdir(self._conf_dir)
	
		self.init()
		
	def sqlite3_connect(self):
		c2padmin.config.node_system_locks['sqlite'].acquire()
		self._db = sqlite3.connect(self._db_path)
		
	def sqlite3_close(self):
		self._db.close()
		c2padmin.config.node_system_locks['sqlite'].release()
	
	def init(self):
		"""docstring for init"""
		self.sqlite3_connect()
		c = self._db.cursor()
		version = c.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='version'").fetchone()
		if(version == None):
			version = 0
			
		else:
			version = c.execute("SELECT version FROM version").fetchone()
			if(version == None):
				version = 0
				
			else:
				version = version[0]
		
		if(version < STRUCTURE_VERSION):
			self.sqlite3_close()
			try:
				self.iso_cache_purge(False)
				
			except:
				pass
				
			os.unlink(self._db_path)
			self.sqlite3_connect()
			c = self._db.cursor()
		
		for q in DB_STRUCTURE:
			c.execute(q)
			
		if(version < STRUCTURE_VERSION):
			c.execute("INSERT INTO version(version) VALUES (?)", (STRUCTURE_VERSION,))
			
		self._db.commit()
		self.sqlite3_close()
		return
	
	def add_iso_record(self, url, size, last_modified, path):
		self.sqlite3_connect()
		c = self._db.cursor()
		c.execute("INSERT INTO isos(url,size,last_modified,path,status) VALUES (?,?,?,?,'OK')",(url,size,last_modified,path))	
		self._db.commit()
		self.sqlite3_close()
		
	def get_iso_record(self, path):
		self.sqlite3_connect()
		c = self._db.cursor()
		record = c.execute("SELECT url,size,last_modified,path,status FROM isos WHERE path = ?", (path,)).fetchone()
		self.sqlite3_close()
		return record
		
	def iso_cache_purge(self, only_incomplete = True):
		self.sqlite3_connect()
		c = self._db.cursor()
		try:
			if(only_incomplete):
				paths = c.execute("SELECT path FROM isos WHERE status != 'OK'").fetchall()
				for path in paths:
					try:
						os.unlink(path[0])
					
					except:
						pass
			
				c.execute("DELETE FROM isos WHERE status != 'OK'")
			
			else:
				paths = c.execute("SELECT path FROM isos").fetchall()
				for path in paths:
					try:
						os.unlink(path[0])
					
					except:
						pass
					
				c.execute("DELETE FROM isos")
				
		except:
			pass
			
		self._db.commit()
		self.sqlite3_close()
		
	def remove_lru_iso(self):
		self.sqlite3_connect()
		c = self._db.cursor()
		iso = c.execute("SELECT * FROM isos WHERE status = 'OK' AND last_used < DATETIME('now', '-5 hour') ORDER BY last_used ASC LIMIT 1").fetchone()
		if(iso != None):
			logger.info('ISO image {} will be removed'.format(iso[4]))
			c.execute('DELETE FROM isos WHERE url=?', (iso[0],))
			self._db.commit()
			self.sqlite3_close()
			os.unlink(iso[4])
		
		else:
			self.sqlite3_close()
	
	def get_isos_cache_size(self):
		self.sqlite3_connect()
		c = self._db.cursor()
		size = c.execute("SELECT SUM(size) FROM isos WHERE status == 'OK'").fetchone()[0]
		self.sqlite3_close()
		return size if(size != None) else 0
		
	def save_iso(self, url = None, fh = None, only_bootable = True, progress = None):
		
		if(url == None and fh == None):
			return False
			
		elif(url != None and fh != None):
			return False
		
		if(url != None):
			r = requests.get(url, stream=True)
			if(r.status_code != 200 or r.headers['content-type'] not in ('application/octet-stream', 'application/x-iso9660-image')):
				r.close()
				if(r.status_code != 200):
					logger.warning('Bad ISO url. Return code {}'.format(r.status_code))
				
				else:
					logger.warning('Bad ISO url. Content type is {}'.format(r.headers['content-type']))
					
				return False
				
			self.sqlite3_connect()
			c = self._db.cursor()
			c.execute("SELECT size,last_modified,path,status FROM isos WHERE url=? LIMIT 1", (url,))
			rec = c.fetchone()
			
			if(rec != None):
				if(rec[3] == 'OK'):
					if(int(r.headers['content-length']) == rec[0] and r.headers.get('last-modified',r.headers['date']) == rec[1]):
						if(os.path.exists(rec[2]) and os.stat(rec[2]).st_size == rec[0]):
							r.close()
							self.sqlite3_close()
							return rec[2]
						
					# something changed. Delete the old file and download the new one
					if(os.path.exists(rec[2])):
						os.unlink(rec[2])
						
					c.execute('DELETE FROM isos WHERE url=?', (url,))
						
				else:
					self.sqlite3_close()
					retVal = rec[2]
					while 1:
						self.sqlite3_connect()
						c = self._db.cursor()
						c.execute("SELECT status FROM isos WHERE url=? LIMIT 1", (url,))
						rec = c.fetchone()
						if(rec == None):
							logger.error('ISO has been deleted')
							self.sqlite3_close()
							return False
							
						if(rec[0] == 'OK'):
							c.execute("UPDATE isos SET last_used = DATETIME('now')")
							self.sqlite3_close()
							return retVal
							
						else:
							if(progress != None):
								progress.set(float(rec[0]))
						
						self.sqlite3_close()	
						time.sleep(1) 
					
			filename = url.split('/')[-1]
			cachepath = os.path.join(self._isos_dir,filename)
			c.execute("INSERT INTO isos(url,size,last_modified,path,status) VALUES (?,?,?,?,'0')",(url,int(r.headers['content-length']),r.headers.get('last-modified',r.headers['date']),cachepath))	
			self._db.commit()
			self.sqlite3_close()
			
			cachefile = open(cachepath,'w')
			
			data_size = int(r.headers['content-length'])
			downloaded = 0
			for chunk in r.iter_content(1<<20):
				cachefile.write(chunk)
				downloaded += 1<<20
				if progress != None:
					progress.set(0.97 * (float(downloaded) / float(data_size)))
				
					self.sqlite3_connect()
					c = self._db.cursor()
					c.execute("UPDATE isos SET status = ? WHERE url=?", (str(0.97 * (float(downloaded) / float(data_size))), url))
					self._db.commit()
					self.sqlite3_close()
			
			r.close()
			cachefile.close()
			
			if(only_bootable):
				try:
					file_type = subprocess.check_output(['/usr/bin/file', cachepath]).split(':')[1].strip()
				
				except:
					logger.warning('Couldn\'t get ISO file type')
					self.sqlite3_close()
					os.unlink(cachepath)
					return False
			
				if(re.match('^.*ISO 9660.*\(bootable\)', file_type) == None):
					logger.warning('Bad ISO file type')
					self.sqlite3_close()
					os.unlink(cachepath)
					return False
			
			self.sqlite3_connect()
			c = self._db.cursor()
			c.execute("UPDATE isos SET status = 'OK', last_used = DATETIME('now') WHERE url=?", (url,))
			self._db.commit()
			self.sqlite3_close()
			return cachepath
			
		# if fh != None:
		# unimplemented
		# TODO
		return False
		
	def save_config(self, name, data):
		self.sqlite3_connect()
		try:
			conf_fh = open(os.path.join(self._conf_dir, name + '.p'), 'w')
			pickle.dump(data, conf_fh)
			conf_fh.close()
			cur = self._db.cursor()
			cur.execute('INSERT OR REPLACE INTO config(name, dirty) VALUES (?, 0)', (name, ))
			self._db.commit()
			
		except:
			self.sqlite3_close()
			return False
		
		self.sqlite3_close()
		return True
		
	def load_config(self,name, commit = False):
		self.sqlite3_connect()
		cur = self._db.cursor()
		cur.execute('SELECT dirty FROM config WHERE name = ?', (name, ))
		res = cur.fetchone()
		if(res == None or res[0] == 1):
			self.sqlite3_close()
			return None
		
		try:
			conf_fh = open(os.path.join(self._conf_dir, name + '.p'), 'r')
			retVal = pickle.load(conf_fh)
			conf_fh.close()
			
			if(commit):
				cur.execute('UPDATE config SET dirty = 1 WHERE name = ?', (name, ))
				self._db.commit()
				
		except:
			self.sqlite3_close()
			return False
		
		self.sqlite3_close()
		return retVal
		
	def commit_config(self, name):
		self.sqlite3_connect()
		cur = self._db.cursor()
		cur.execute('UPDATE config SET dirty = 1 WHERE name = ?', (name, ))
		self._db.commit()
		self.sqlite3_close()
		
if __name__ == '__main__':
	c = LocalCache()
	#print c.save_iso('http://mirror.hosting90.cz/debian-cd/current/amd64/iso-cd/debian-9.1.0-amd64-netinst.iso')
	print c.save_iso('http://ftp-stahuj.centrum.cz/dl/ea950b60b6376b8f7f1b2489dbd397e5/59d233df/stahuj/download/software/secured/7/7zip/1700/64-bit/7z1700-x64.exe')
	
	